@@ -9,6 +9,7 @@ from . import display
 from . import gitcmds
 from . import icons
 from . import qtutils
+from . import resources
 from .i18n import N_
 from .interaction import Interaction
 from .widgets import completion
@@ -19,6 +20,13 @@ from .widgets.selectcommits import select_commits
 from .widgets.selectcommits import select_commits_and_output
 
 
+def copy_commit_id_to_clipboard(context):
+    """Copy the current commit ID to the clipboard"""
+    status, commit_id, _ = context.git.rev_parse('HEAD')
+    if status == 0 and commit_id:
+        qtutils.set_clipboard(commit_id)
+
+
 def delete_branch(context):
     """Launch the 'Delete Branch' dialog."""
     icon = icons.discard()
@@ -339,26 +347,32 @@ def reset_soft(context):
     """Run "git reset --soft" to reset the branch HEAD"""
     title = N_('Reset Branch (Soft)')
     ok_text = N_('Reset Branch')
-    ref = choose_ref(context, title, ok_text, default='HEAD^')
+    default = context.settings.get_value('reset::soft', 'ref', default='HEAD^')
+    ref = choose_ref(context, title, ok_text, default=default)
     if ref:
         cmds.do(cmds.ResetSoft, context, ref)
+        context.settings.set_value('reset::soft', 'ref', ref)
 
 
 def reset_mixed(context):
     """Run "git reset --mixed" to reset the branch HEAD and staging area"""
     title = N_('Reset Branch and Stage (Mixed)')
     ok_text = N_('Reset')
-    ref = choose_ref(context, title, ok_text, default='HEAD^')
+    default = context.settings.get_value('reset::mixed', 'ref', default='HEAD^')
+    ref = choose_ref(context, title, ok_text, default=default)
     if ref:
         cmds.do(cmds.ResetMixed, context, ref)
+        context.settings.set_value('reset::mixed', 'ref', ref)
 
 
 def reset_keep(context):
     """Run "git reset --keep" safe reset to avoid clobbering local changes"""
     title = N_('Reset All (Keep Unstaged Changes)')
-    ref = choose_ref(context, title, N_('Reset and Restore'))
+    default = context.settings.get_value('reset::keep', 'ref', default='HEAD^')
+    ref = choose_ref(context, title, N_('Reset and Restore'), default=default)
     if ref:
         cmds.do(cmds.ResetKeep, context, ref)
+        context.settings.set_value('reset::keep', 'ref', ref)
 
 
 def reset_merge(context):
@@ -370,27 +384,92 @@ def reset_merge(context):
     """
     title = N_('Restore Worktree and Reset All (Merge)')
     ok_text = N_('Reset and Restore')
-    ref = choose_ref(context, title, ok_text, default='HEAD^')
+    default = context.settings.get_value('reset::merge', 'ref', default='HEAD^')
+    ref = choose_ref(context, title, ok_text, default=default)
     if ref:
         cmds.do(cmds.ResetMerge, context, ref)
+        context.settings.set_value('reset::merge', 'ref', ref)
 
 
 def reset_hard(context):
     """Run "git reset --hard" to fully reset the working tree and staging area"""
     title = N_('Restore Worktree and Reset All (Hard)')
     ok_text = N_('Reset and Restore')
-    ref = choose_ref(context, title, ok_text, default='HEAD^')
+    default = context.settings.get_value('reset::hard', 'ref', default='HEAD^')
+    ref = choose_ref(context, title, ok_text, default=default)
     if ref:
         cmds.do(cmds.ResetHard, context, ref)
+        context.settings.set_value('reset::hard', 'ref', ref)
 
 
 def restore_worktree(context):
     """Restore the worktree to the content from the specified commit"""
     title = N_('Restore Worktree')
     ok_text = N_('Restore Worktree')
-    ref = choose_ref(context, title, ok_text, default='HEAD^')
+    default = context.settings.get_value('restore::worktree', 'ref', default='HEAD^')
+    ref = choose_ref(context, title, ok_text, default=default)
     if ref:
         cmds.do(cmds.RestoreWorktree, context, ref)
+        context.settings.set_value('restore::worktree', 'ref', ref)
+
+
+def build_layout_menu(widget, menu):
+    """Add layouts from ~/.config/git-cola/layouts to the specified menu"""
+    directory = resources.xdg_config_home('git-cola', 'layouts')
+    if os.path.isdir(directory):
+        layouts = sorted(os.listdir(directory))
+    else:
+        layouts = []
+    suffix = '.layout'
+    if layouts:
+        menu.addSeparator()
+    for layout in layouts:
+        if layout.endswith(suffix):
+            layout_name = layout[: -len(suffix)]
+        else:
+            layout_name = layout
+        layout_filename = os.path.join(directory, layout)
+        load_layout_action = qtutils.add_action(
+            widget,
+            layout_name,
+            lambda filename=layout_filename: load_layout_file(widget, filename),
+        )
+        menu.addAction(load_layout_action)
+
+
+def save_layout(widget):
+    """Save the current widget layout to a file"""
+    default_filename = resources.xdg_config_home(
+        'git-cola', 'layouts', 'default.layout'
+    )
+    parent_dir = os.path.dirname(default_filename)
+    if not os.path.isdir(parent_dir):
+        os.makedirs(parent_dir)
+    filename = qtutils.save_as(default_filename)
+    if not filename:
+        return
+    state = widget.layout_state()
+    with open(filename, 'wb') as output:
+        output.write(state)
+
+
+def load_layout(widget):
+    """Choose a Qt layout file and apply it to the current widget"""
+    directory = resources.xdg_config_home('git-cola', 'layouts')
+    if not os.path.isdir(directory):
+        os.makedirs(directory)
+    filename = qtutils.existing_file(directory, title=N_('Load Layout'))
+    load_layout_file(widget, filename)
+
+
+def load_layout_file(widget, filename):
+    """Load a Qt layout file into the specified widget"""
+    if not filename or not os.path.isfile(filename):
+        return
+    with open(filename, 'rb') as handle:
+        state = handle.read()
+    if state:
+        widget.apply_layout(state)
 
 
 def install():
